"""Manage reference image and ROIs."""

import abc
import dataclasses
from typing import TYPE_CHECKING, Generic, TypeVar

import cv2
import numpy as np
import numpy.typing as npt

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

__all__ = [
    "DataTypeVar",
    "DynamicROI",
    "StaticROI",
    "sanitize_ROI",
    "ReferenceBase",
    "ReferenceData",
    "Reference",
]


DataTypeVar = TypeVar("DataTypeVar", bound="DataclassInstance")
"""Type variable for :attr:`ReferenceBase.DataType`."""


DynamicROI = tuple[int | None, int | None, int | None, int | None]
"""Type annotation for ROI whose upper limits can be dynamically determined.

This is a tuple of ``(x0, y0, x1, y1)``, where items can be integer or
:obj:`None` by Python's slicing convention.
"""

StaticROI = tuple[int, int, int, int]
"""Type annotation for ROI whose items are all static.

This is a tuple of ``(x0, y0, x1, y1)``, where every item is
nonnegative :class:`int`.
"""


def sanitize_ROI(roi: DynamicROI, h: int, w: int) -> StaticROI:
    """Convert dynamic ROI to static ROI.

    Arguments:
        roi: Tuple in ``(x0, y0, x1, y1)``.
            Items can be integer or :obj:`None` by Python's
            slicing convention.
        h, w: Height and width of the image.

    Returns:
        Tuple in ``(x0, y0, x1, y1)``. Values are converted to
        positive integers.
    """
    full_roi = (0, 0, w, h)
    max_vars = (w, h, w, h)

    ret = list(roi)
    for i, var in enumerate(roi):
        if var is None:
            ret[i] = full_roi[i]
        elif var < 0:
            ret[i] = max_vars[i] + var
    return tuple(ret)  # type: ignore[return-value]


class ReferenceBase(abc.ABC, Generic[DataTypeVar]):
    """Abstract base class for reference object.

    Reference object stores reference image, which is a binary image of uncoated
    substrate. It also contains ROIs for template region and substrate region in
    the reference image.

    External API can use the following members to get analysis results of
    concrete subclasses.

    * :attr:`DataType`: Dataclass type for the analysis result.
    * :meth:`analyze`: :attr:`DataType` instance containing analysis result.
    * :meth:`draw`: Visualized result.

    Arguments:
        image: Binary reference image.
    """

    DataType: type[DataTypeVar]
    """Return type of :attr:`analyze`.

    Concrete subclass must assign this attribute with dataclass type.
    """

    def __init__(self, image: npt.NDArray[np.uint8]):
        """Initialize the instance.

        *image* is set to be not writable.
        """
        self._image = image
        self._image.setflags(write=False)

    @property
    def image(self) -> npt.NDArray[np.uint8]:
        """Binary reference image.

        For immutability, this image is not writable.
        """
        return self._image

    @property
    @abc.abstractmethod
    def templateROI(self) -> StaticROI:
        """ROI for template image."""

    @property
    @abc.abstractmethod
    def substrateROI(self) -> StaticROI:
        """ROI for substrate image."""

    @abc.abstractmethod
    def analyze(self) -> DataTypeVar:
        """Return analysis result as dataclass.

        Return type must be :attr:`DataType`.
        """

    def draw(
        self,
        templateColor: tuple[int, int, int] = (255, 0, 0),
        templateThickness: int = 1,
        substrateColor: tuple[int, int, int] = (0, 255, 0),
        substrateThickness: int = 1,
    ) -> npt.NDArray[np.uint8]:
        """Return visualization result in RGB format.

        Arguments:
            templateColor: Template ROI box color for :func:`cv2.rectangle`.
            templateThickness: Template ROI box line width for :func:`cv2.rectangle`.
            substrateColor: Substrate ROI box color for :func:`cv2.rectangle`.
            substrateThickness: Substrate ROI box line width for :func:`cv2.rectangle`.
        """
        ret = cv2.cvtColor(self.image, cv2.COLOR_GRAY2RGB)

        if templateThickness > 0:
            x0, y0, x1, y1 = self.templateROI
            cv2.rectangle(ret, (x0, y0), (x1, y1), templateColor, templateThickness)

        if substrateThickness > 0:
            x0, y0, x1, y1 = self.substrateROI
            cv2.rectangle(ret, (x0, y0), (x1, y1), substrateColor, substrateThickness)

        return ret  # type: ignore[return-value]


@dataclasses.dataclass
class ReferenceData:
    """Analysis data for :class:`Reference`."""

    pass


class Reference(ReferenceBase[ReferenceData]):
    """Reference image with ROIs specified.

    Arguments:
        image: Binary reference image.
        templateROI: ROI for template image.
        substrateROI: ROI for substrate image.

    Examples:
        .. plot::
            :include-source:
            :context: reset

            >>> import cv2
            >>> from finitedepth import get_sample_path, Reference
            >>> img = cv2.imread(get_sample_path("ref.png"), cv2.IMREAD_GRAYSCALE)
            >>> _, bin = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
            >>> ref = Reference(bin, (10, 10, 1250, 200), (100, 100, 1200, 500))
            >>> import matplotlib.pyplot as plt #doctest: +SKIP
            >>> plt.imshow(ref.draw()) #doctest: +SKIP
    """

    DataType = ReferenceData
    """Return :obj:`ReferenceData`."""

    def __init__(
        self,
        image: npt.NDArray[np.uint8],
        templateROI: DynamicROI = (0, 0, None, None),
        substrateROI: DynamicROI = (0, 0, None, None),
    ):
        """Initialize the instance.

        *templateROI* and *substrateROI* are converted to :obj:`StaticROI` using
        :func:`sanitize_ROI`.
        """
        super().__init__(image)

        h, w = image.shape[:2]
        self._templateROI = sanitize_ROI(templateROI, h, w)
        self._substrateROI = sanitize_ROI(substrateROI, h, w)

    @property
    def templateROI(self) -> StaticROI:
        """ROI for template image."""
        return self._templateROI

    def analyze(self):
        """Return empty :class:`ReferenceData`."""
        return self.DataType()

    @property
    def substrateROI(self) -> StaticROI:
        """ROI for substrate image."""
        return self._substrateROI



"""Analyze coating layer and similarity measures [#senin]_ [#sim]_.

.. [#senin] P., Senin (2008)
.. [#sim] https://pypi.org/project/similaritymeasures/
"""

# TODO: fix docstring formats

import abc
import dataclasses
from typing import TYPE_CHECKING, Generic, TypeVar

import cv2
import numpy as np
import numpy.typing as npt
from numba import njit  # type: ignore
from scipy.optimize import root  # type: ignore
from scipy.spatial.distance import cdist  # type: ignore
from shapely import LineString, offset_curve  # type: ignore

from .cache import attrcache
from .substrate import RectSubstrate, SubstrateBase

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

__all__ = [
    "CoatingLayerBase",
    "CoatingLayer",
    "RectLayerShape",
    "images_XOR",
    "images_ANDXOR",
    "equidistant_interpolate",
    "parallel_curve",
    "acm",
    "owp",
]


SubstTypeVar = TypeVar("SubstTypeVar", bound=SubstrateBase)
"""Type variable for the substrate type of :class:`CoatingLayerBase`."""
DataTypeVar = TypeVar("DataTypeVar", bound="DataclassInstance")
"""Type variable for :attr:`CoatingLayerBase.DataType`."""


class CoatingLayerBase(abc.ABC, Generic[SubstTypeVar, DataTypeVar]):
    """Abstract base class for coating layer object.

    Coating layer object stores :class:`SubstrateBase` object and target image, which is
    a binary image of coated substrate. The role of coating layer object is to acquire
    coating layer region by template matching and analyze its shape.

    External API can use the following members to get analysis results of
    concrete subclasses.

    * :attr:`DataType`: Dataclass type for the analysis result.
    * :meth:`analyze`: :attr:`DataType` instance containing analysis result.
    * :meth:`draw`: Visualized result.

    Arguments:
        image: Binary target image.
        substrate: Substrate instance storing binary reference image.
        tempmatch: Pre-computed template matching result.
            External constructor can pass this argument to force the template matching
            result. If not passed, :meth:`match_template` performs matching.
    """

    DataType: type[DataTypeVar]
    """Return type of :attr:`analyze`.

    Concrete subclass must assign this attribute with dataclass type.
    """

    def __init__(
        self,
        image: npt.NDArray[np.uint8],
        substrate: SubstTypeVar,
        *,
        tempmatch: tuple[tuple[int, ...], float] | None = None,
    ):
        """Initialize the instance.

        *image* is set to be immutable, and template matching is performed if
        *tempmatch* is ``None``.
        """
        self._image = image
        self._image.setflags(write=False)
        self._substrate = substrate

        if tempmatch is None:
            image = self.image
            x0, y0, x1, y1 = self.substrate.reference.templateROI
            template = self.substrate.reference.image[y0:y1, x0:x1]
            self._tempmatch = self.match_template(image, template)
        else:
            self._tempmatch = tempmatch

    @property
    def image(self) -> npt.NDArray[np.uint8]:
        """Binary target image.

        For immutability, this image is not writable.
        """
        return self._image

    @property
    def substrate(self) -> SubstTypeVar:
        """Substrate instance which contains substrate and reference information."""
        return self._substrate

    @property
    def tempmatch(self) -> tuple[tuple[int, ...], float]:
        """Template matching location and score."""
        return self._tempmatch

    def match_template(
        self, image: npt.NDArray[np.uint8], template: npt.NDArray[np.uint8]
    ) -> tuple[tuple[int, ...], float]:
        """Perform template matching between *image* and *template*.

        Template matching is performed using :func:`cv2.matchTemplate` with
        :obj:`cv2.TM_SQDIFF_NORMED`. Subclass may override this method to apply
        other algorithm.

        Arguments:
            image: Binary target image.
            template: Binary template image.
        """
        res = cv2.matchTemplate(image, template, cv2.TM_SQDIFF_NORMED)
        score, _, loc, _ = cv2.minMaxLoc(res)
        return (tuple(loc), score)

    def substrate_point(self) -> npt.NDArray[np.int32]:
        """Upper left point of the substrate image in target image.

        Returns:
            Coordinates in ``(x, y)``.
        """
        temp_point, _ = self.tempmatch
        t = self.substrate.reference.templateROI[:2]
        s = self.substrate.reference.substrateROI[:2]
        return np.array(temp_point, dtype=np.int32) - t + s

    @attrcache("_coated_substrate")
    def coated_substrate(self) -> npt.NDArray[np.bool_]:
        """Coated substrate region.

        Returns:
            Target image without artifacts, e.g., bath surface.
        """
        _, img = cv2.connectedComponents(cv2.bitwise_not(self.image))
        x, y = (self.substrate_point() + self.substrate.region_points()).T
        return np.isin(img, img[y, x])

    @attrcache("_extracted_layer")
    def extract_layer(self) -> npt.NDArray[np.bool_]:
        """Coating layer region extracted from target image."""
        # remove the substrate
        x0, y0 = self.substrate_point()
        subst_mask = self.substrate.regions() >= 0
        ret = images_ANDXOR(self.coated_substrate(), subst_mask, (x0, y0))
        ret[:y0, :] = False
        return ret

    @abc.abstractmethod
    def analyze(self) -> DataTypeVar:
        """Return analysis result as dataclass.

        Return type must be :attr:`DataType`.
        """

    @abc.abstractmethod
    def draw(self, *args, **kwargs) -> npt.NDArray[np.uint8]:
        """Return visualization result."""


@dataclasses.dataclass
class CoatingLayerData:
    """Analysis data for :class:`CoatingLayer`."""

    pass


class CoatingLayer(CoatingLayerBase[SubstrateBase, CoatingLayerData]):
    """Basic implementation of coating layer without any analysis.

    Arguments:
        image: Binary target image.
        substrate: Substrate instance.
        tempmatch: Pre-computed template matching result.

    Examples:
        Construct substrate instance first.

        .. plot::
            :include-source:
            :context: reset

            >>> import cv2
            >>> from finitedepth import get_sample_path, Reference, Substrate
            >>> img = cv2.imread(get_sample_path("ref.png"), cv2.IMREAD_GRAYSCALE)
            >>> _, bin = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
            >>> ref = Reference(bin, (10, 10, 1250, 200), (100, 100, 1200, 500))
            >>> subst = Substrate(ref)
            >>> import matplotlib.pyplot as plt #doctest: +SKIP
            >>> plt.imshow(subst.draw()) #doctest: +SKIP

        Then, construct coating layer instance.

        .. plot::
            :include-source:
            :context: close-figs

            >>> from finitedepth import CoatingLayer
            >>> img = cv2.imread(get_sample_path("coat.png"), cv2.IMREAD_GRAYSCALE)
            >>> _, bin = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
            >>> coat = CoatingLayer(bin, subst)
            >>> plt.imshow(coat.draw()) #doctest: +SKIP
    """

    DataType = CoatingLayerData
    """Return :obj:`CoatingLayerData`."""

    def analyze(self):
        """Return empty :class:`CoatingLayerData`."""
        return self.DataType()

    def draw(
        self,
        subtraction_mode: str = "none",
        layer_color: tuple[int, int, int] = (255, 0, 0),
        layer_thickness: int = -1,
    ) -> npt.NDArray[np.uint8]:
        """Subtract the template match result and paint the coating layer.

        Arguments:
            subtraction_mode ({`'none', 'template', 'substrate', 'full'`}): Subtraction
                mode. `'template'` and `'substrate'` removes overlapping template region
                and substrate region, respectively. `'full'` removes both.
            layer_color: Layer color for :func:`cv2.drawContours`.
            layer_thickness: Layer thickness for :func:`cv2.drawContours`.
        """
        image = cv2.cvtColor(self.image, cv2.COLOR_GRAY2RGB)

        if subtraction_mode not in ["none", "template", "substrate", "full"]:
            raise TypeError("Unrecognized subtraction mode: %s" % subtraction_mode)
        if subtraction_mode in ["template", "full"]:
            x0, y0, x1, y1 = self.substrate.reference.templateROI
            tempImg = self.substrate.reference.image[y0:y1, x0:x1]
            h, w = tempImg.shape[:2]
            (X0, Y0), _ = self.tempmatch
            binImg = self.image[Y0 : Y0 + h, X0 : X0 + w]
            mask = images_XOR(~binImg.astype(bool), ~tempImg.astype(bool))
            image[Y0 : Y0 + h, X0 : X0 + w][~mask] = 255
        if subtraction_mode in ["substrate", "full"]:
            x0, y0, x1, y1 = self.substrate.reference.substrateROI
            substImg = self.substrate.reference.image[y0:y1, x0:x1]
            h, w = substImg.shape[:2]
            X0, Y0 = self.substrate_point()
            binImg = self.image[Y0 : Y0 + h, X0 : X0 + w]
            mask = images_XOR(~binImg.astype(bool), ~substImg.astype(bool))
            image[Y0 : Y0 + h, X0 : X0 + w][~mask] = 255

        cnts, _ = cv2.findContours(
            self.extract_layer().astype(np.uint8),
            cv2.RETR_EXTERNAL,
            cv2.CHAIN_APPROX_NONE,
        )
        cv2.drawContours(image, cnts, -1, layer_color, layer_thickness)

        return image  # type: ignore[return-value]


@dataclasses.dataclass
class RectLayerShapeData:
    """Analysis data for :class:`RectLayerShape`.

    Arguments:
        LayerLength_Left, LayerLength_Right: Length of the layer on each wall.
        Conformality: Conformality of the coating layer.
        AverageThickness: Average thickness of the coating layer.
        Roughness: Roughness of the coating layer.
        MaxThickness_Left, MaxThickness_Bottom, MaxThickness_Right: Regional maximum
            thicknesses.
        MatchError: Template matching error between ``0`` to ``1``.
            ``0`` means perfect match.
    """

    LayerLength_Left: np.float64
    LayerLength_Right: np.float64

    Conformality: float
    AverageThickness: np.float64
    Roughness: float

    MaxThickness_Left: np.float64
    MaxThickness_Bottom: np.float64
    MaxThickness_Right: np.float64

    MatchError: float


class RectLayerShape(CoatingLayerBase[RectSubstrate, RectLayerShapeData]):
    """Coating layer over rectangular substrate.

    Arguments:
        image: Binary target image.
        substrate: Substrate instance.
        opening_ksize: Kernel size for morphological operation.
            Elements must be zero or odd number.
        reconstruct_radius: Radius of the "safe zone" for noise removal.
            Imaginary circles with this radius are drawn on bottom corners of the
            substrate. Connected components not passing these circles are regarded as
            image artifacts.
        roughness_measure ({`'DTW', 'SDTW'`}): Similarity measure to quantify roughness.
            `'DTW'` is dynamic time warping and `'SDTW'` is its root mean square.
        tempmatch: Pre-computed template matching result.

    Examples:
        Construct substrate instance first.

        .. plot::
            :include-source:
            :context: reset

            >>> import cv2
            >>> from finitedepth import get_sample_path, Reference, RectSubstrate
            >>> img = cv2.imread(get_sample_path("ref.png"), cv2.IMREAD_GRAYSCALE)
            >>> _, bin = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
            >>> ref = Reference(bin, (10, 10, 1250, 200), (100, 100, 1200, 500))
            >>> subst = RectSubstrate(ref, 3.0, 1.0, 0.01)
            >>> import matplotlib.pyplot as plt #doctest: +SKIP
            >>> plt.imshow(subst.draw()) #doctest: +SKIP

        Then, construct coating layer instance.

        .. plot::
            :include-source:
            :context: close-figs

            >>> from finitedepth import RectLayerShape
            >>> img = cv2.imread(get_sample_path("coat.png"), cv2.IMREAD_GRAYSCALE)
            >>> _, bin = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
            >>> coat = RectLayerShape(bin, subst, (1, 1), 50, "DTW")
            >>> plt.imshow(
            ...     coat.draw(conformality_step=10, roughness_step=10)
            ... ) #doctest: +SKIP
    """

    DataType = RectLayerShapeData
    """Return :obj:`RectLayerShapeData`."""

    def __init__(
        self,
        image: npt.NDArray[np.uint8],
        substrate: RectSubstrate,
        opening_ksize: tuple[int, int],
        reconstruct_radius: int,
        roughness_measure: str,
        *,
        tempmatch: tuple[tuple[int, ...], float] | None = None,
    ):
        """Initialize the instance.

        Check whether kernel size are zero or odd and roughness measure is either
        `'DTW'` or `'SDTW'`.
        """
        if not all(i == 0 or (i > 0 and i % 2 == 1) for i in opening_ksize):
            raise ValueError("Kernel size must be zero or odd.")
        if roughness_measure not in ["DTW", "SDTW"]:
            raise TypeError(f"Unknown roughness measure: {roughness_measure}")
        super().__init__(image, substrate, tempmatch=tempmatch)
        self._opening_ksize = opening_ksize
        self._reconstruct_radius = reconstruct_radius
        self._roughness_measure = roughness_measure

    @property
    def opening_ksize(self) -> tuple[int, int]:
        """Kernel size for morphological operation."""
        return self._opening_ksize

    @property
    def reconstruct_radius(self) -> int:
        """Radius of the "safe zone" for noise removal."""
        return self._reconstruct_radius

    @property
    def roughness_measure(self) -> str:
        """Similarity measure to quantify roughness."""
        return self._roughness_measure

    @attrcache("_layer_contours")
    def layer_contours(self) -> tuple[npt.NDArray[np.int32], ...]:
        """Find contours of coating layer region.

        This method finds external contours of :meth:`~.extract_layer`.
        Each contour encloses each discrete region of coating layer.
        """
        layer_cnts, _ = cv2.findContours(
            self.extract_layer().astype(np.uint8),
            cv2.RETR_EXTERNAL,
            cv2.CHAIN_APPROX_NONE,
        )
        return tuple(layer_cnts)

    @attrcache("_interfaces")
    def interfaces(self) -> tuple[npt.NDArray[np.int64], ...]:
        """Find solid-liquid interfaces.

        A substrate can have contact with multiple discrete coating layer regions,
        and a single coating layer region can have multiple contacts to the substrate.
        This method returns indices for :meth:`~finitedepth.PolySubstrateBase.contour`
        where solid-liquid interfaces start and stop.

        Returns:
            tuple of arrays. ``i``-th array represents ``i``-th coating layer region in
            :meth:`layer_contours`. Shape of the array is ``(N, 2)``, where ``N`` is the
            number of contacts the coating layer region makes. Each column represents
            starting and ending indices for the interface interval in substrate contour.

        Note:
            Each interval describes continuous patch on the substrate contour covered
            by the layer. To acquire the interface points, slice :attr:`substrate`'s
            :meth:`~finitedepth.PolySubstrateBase.contour` with the indices.
        """
        subst_cnt = self.substrate.contour() + self.substrate_point()
        ret = []
        for layer_cnt in self.layer_contours():
            H, W = self.image.shape[:2]
            lcnt_img = np.zeros((H, W), dtype=np.uint8)
            lcnt_img[layer_cnt[..., 1], layer_cnt[..., 0]] = 255
            dilated_lcnt = cv2.dilate(lcnt_img, np.ones((3, 3))).astype(bool)

            x, y = subst_cnt.transpose(2, 0, 1)
            mask = dilated_lcnt[np.clip(y, 0, H - 1), np.clip(x, 0, W - 1)]

            # Find indices of continuous True blocks
            idxs = np.where(
                np.diff(np.concatenate(([False], mask[:, 0], [False]))) == 1
            )[0].reshape(-1, 2)
            ret.append(idxs)
        return tuple(ret)

    @attrcache("_contour")
    def contour(self) -> npt.NDArray[np.int32]:
        """Contour of the entire coated substrate."""
        (cnt,), _ = cv2.findContours(
            self.coated_substrate().astype(np.uint8),
            cv2.RETR_EXTERNAL,
            cv2.CHAIN_APPROX_NONE,
        )
        return cnt

    def capbridge_broken(self) -> bool:
        """Check if capillary bridge is ruptured.

        As substrate is withdrawn from fluid bath, capillary bridge forms between the
        coating layer and bulk fluid and then ruptures.
        """
        p0 = self.substrate_point()
        _, bl, br, _ = self.substrate.contour()[self.substrate.vertices()]
        (B,) = p0 + bl
        (C,) = p0 + br
        top = np.max([B[1], C[1]])
        bot = self.image.shape[0]
        if top > bot:
            # substrate is located outside of the frame
            return False
        left = B[0]
        right = C[0]
        roi_binimg = self.image[top:bot, left:right]
        return bool(np.any(np.all(roi_binimg, axis=1)))

    @attrcache("_extracted_layer")
    def extract_layer(self) -> npt.NDArray[np.bool_]:
        """Coating layer region extracted from target image.

        Error pixels are removed by performing morphological operation and removing the
        disconnected components that are far from the substrate.
        """
        # Perform opening to remove error pixels. We named the parameter as
        # "closing" because the coating layer is black in original image, but
        # in fact we do opening since the layer is True in extracted layer.
        ksize = self.opening_ksize
        if any(i == 0 for i in ksize):
            img = super().extract_layer().astype(np.uint8) * 255
        else:
            kernel = cv2.getStructuringElement(cv2.MORPH_RECT, ksize)
            img = cv2.morphologyEx(
                super().extract_layer().astype(np.uint8) * 255,
                cv2.MORPH_OPEN,
                kernel,
            )

        # closed image may still have error pixels, and at least we have to
        # remove the errors that are disconnected to the layer.
        # we identify the layer pixels as the connected components that are
        # close to the lower vertices.
        vicinity_mask = np.zeros(img.shape, np.uint8)
        p0 = self.substrate_point()
        _, bl, br, _ = self.substrate.contour()[self.substrate.vertices()]
        (B,) = p0 + bl
        (C,) = p0 + br
        R = self.reconstruct_radius
        cv2.circle(
            vicinity_mask, B.astype(np.int32), R, 1, -1
        )  # type: ignore[call-overload]
        cv2.circle(
            vicinity_mask, C.astype(np.int32), R, 1, -1
        )  # type: ignore[call-overload]
        n = np.dot((C - B) / np.linalg.norm((C - B)), np.array([[0, 1], [-1, 0]]))
        pts = np.stack([B, B + R * n, C + R * n, C]).astype(np.int32)
        cv2.fillPoly(vicinity_mask, [pts], 1)  # type: ignore[call-overload]
        _, labels = cv2.connectedComponents(img)
        layer_comps = np.unique(labels[np.where(vicinity_mask.astype(bool))])
        layer_mask = np.isin(labels, layer_comps[layer_comps != 0])

        return layer_mask

    @attrcache("_surface")
    def surface(self) -> tuple[np.int64, np.int64]:
        """Liquid-gas interface of the coating layer.

        Substrate surface exposed to air is considered to be covered by coating layer
        with zero thickness.

        Returns:
            Starting and ending indices for the surface interval in coated substrate
            contour.

        Note:
            To acquire the surface points, slice
            :meth:`~finitedepth.PolySubstrateBase.contour` with the indices.
        """
        if not self.interfaces():
            return (np.int64(-1), np.int64(0))

        (i0, i1) = np.sort(np.concatenate(self.interfaces()).flatten())[[0, -1]]
        subst_cnt = self.substrate.contour() + self.substrate_point()
        endpoints = subst_cnt[[i0, i1]]

        vec = self.contour() - endpoints.transpose(1, 0, 2)
        (I0, I1) = np.argmin(np.linalg.norm(vec, axis=-1), axis=0)
        return (I0, I1 + 1)

    @attrcache("_uniform_layer")
    def uniform_layer(self) -> tuple[np.float64, npt.NDArray[np.float64]]:
        """Imaginary uniform layer.

        Uniform layer is a parallel curve [#parallel-curve]_ of substrate surface
        which has the same cross-sectional area as the actual coating layer.

        Returns:
            Thickness and polyline vertices of the uniform layer.

        .. [#parallel-curve] https://en.wikipedia.org/wiki/Parallel_curve
        """
        if not self.interfaces():
            return (np.float64(0), np.empty((0, 1, 2), np.float64))

        (i0, i1) = np.sort(np.concatenate(self.interfaces()).flatten())[[0, -1]]
        subst_cnt = self.substrate.contour() + self.substrate_point()
        s = subst_cnt[i0:i1]

        A = np.count_nonzero(self.extract_layer())
        (t,) = root(
            lambda x: cv2.contourArea(
                np.concatenate([s, np.flip(parallel_curve(s, x[0]), axis=0)]).astype(
                    np.float32
                )
            )
            - A,
            [0],
        ).x
        return (t, parallel_curve(s, t))

    @attrcache("_conformality")
    def conformality(self) -> tuple[float, npt.NDArray[np.int32]]:
        """DTW-based conformality of the coating layer.

        Returns:
            Conformality and optimal pairs between substrate surface and layer surface.
        """
        if not self.interfaces():
            return (np.nan, np.empty((0, 2), dtype=np.int32))

        (i0, i1) = np.sort(np.concatenate(self.interfaces()).flatten())[[0, -1]]
        subst_cnt = self.substrate.contour() + self.substrate_point()
        intf = subst_cnt[i0:i1]

        I0, I1 = self.surface()
        surf = self.contour()[I0:I1]

        dist = cdist(np.squeeze(surf, axis=1), np.squeeze(intf, axis=1))
        mat = acm(dist)
        path = owp(mat)
        d = dist[path[:, 0], path[:, 1]]
        d_avrg = mat[-1, -1] / len(path)
        C = 1 - np.sum(np.abs(d - d_avrg)) / mat[-1, -1]
        return (float(C), path)

    @attrcache("_roughness")
    def roughness(self) -> tuple[float, npt.NDArray[np.int32]]:
        """DTW-based surface roughness of the coating layer.

        Returns:
            Roughness and optimal pairs between layer surface and uniform layer.
        """
        I0, I1 = self.surface()
        surf = self.contour()[I0:I1]
        _, ul = self.uniform_layer()

        if surf.size == 0 or ul.size == 0:
            return (np.nan, np.empty((0, 2), dtype=np.int32))

        if self.roughness_measure == "DTW":
            ul_len = np.ceil(cv2.arcLength(ul.astype(np.float32), closed=False))
            ul = equidistant_interpolate(ul, int(ul_len))
            dist = cdist(np.squeeze(surf, axis=1), np.squeeze(ul, axis=1))
            mat = acm(dist)
            path = owp(mat)
            roughness = mat[-1, -1] / len(path)
        elif self.roughness_measure == "SDTW":
            ul_len = np.ceil(cv2.arcLength(ul.astype(np.float32), closed=False))
            ul = equidistant_interpolate(ul, int(ul_len))
            dist = cdist(np.squeeze(surf, axis=1), np.squeeze(ul, axis=1))
            mat = acm(dist**2)
            path = owp(mat)
            roughness = np.sqrt(mat[-1, -1] / len(path))
        else:
            path = np.empty((0, 2), dtype=np.int32)
            roughness = -1
        return (float(roughness), path)

    @attrcache("_max_thickness")
    def max_thickness(self) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Regional maximum thicknesses.

        Coating layer is segmented using :meth:`RectSubstrate.sideline_intersections`.
        Points on layer surface and sideline for maximum distance in each region are
        found.

        Returns:
            tuple of two arrays. The first array contains maximum thickness values on
            left, bottom, and right region. Value of ``0`` indicates no coating layer on
            that region. The second array contains points on layer surface and substrate
            lines for the maximum thickness.
            Shape of the array is ``(3, 2, 2)``; 1st axis indicates left, bottom
            and right region, 2nd axis indicates layer surface and substrate line,
            and 3rd axis indicates ``(x, y)`` coordinates.
        """
        corners = self.substrate.sideline_intersections() + self.substrate_point()
        I0, I1 = self.surface()
        surface = self.contour()[I0:I1]
        thicknesses, points = [], []
        for A, B in zip(corners[:-1], corners[1:]):
            AB = B - A
            mask = np.cross(AB, surface - A) >= 0
            pts = surface[mask]
            if pts.size == 0:
                thicknesses.append(np.float64(0))
                points.append(np.array([[-1, -1], [-1, -1]], np.float64))
            else:
                Ap = pts - A
                Proj = A + AB * (np.dot(Ap, AB) / np.dot(AB, AB))[..., np.newaxis]
                dist = np.linalg.norm(Proj - pts, axis=-1)
                max_idx = np.argmax(dist)
                thicknesses.append(dist[max_idx])
                points.append(np.stack([pts[max_idx], Proj[max_idx]]))
        return (np.array(thicknesses), np.array(points))

    def analyze(self):
        """Return :class:`RectLayerShapeData`."""
        _, B, C, _ = self.substrate.sideline_intersections() + self.substrate_point()

        if not self.interfaces():
            LEN_L = LEN_R = np.float64(0)
        else:
            (i0, i1) = np.sort(np.concatenate(self.interfaces()).flatten())[[0, -1]]
            subst_cnt = self.substrate.contour() + self.substrate_point()
            pts = subst_cnt[[i0, i1]]

            Bp = pts - B
            BC = C - B
            Proj = B + BC * (np.dot(Bp, BC) / np.dot(BC, BC))[..., np.newaxis]
            dists = np.linalg.norm(Proj - pts, axis=-1)
            (LEN_L,), (LEN_R,) = dists.astype(np.float64)

        C, _ = self.conformality()
        AVRGTHCK, _ = self.uniform_layer()
        ROUGH, _ = self.roughness()
        (THCK_L, THCK_B, THCK_R), _ = self.max_thickness()

        _, ERR = self.tempmatch

        return self.DataType(
            LEN_L,
            LEN_R,
            C,
            AVRGTHCK,
            ROUGH,
            THCK_L,
            THCK_B,
            THCK_R,
            ERR,
        )

    def draw(
        self,
        background_mode: str = "image",
        subtraction_mode: str = "none",
        layer_color: tuple[int, int, int] = (255, 0, 0),
        layer_thickness: int = -1,
        contactline_color: tuple[int, int, int] = (0, 255, 0),
        contactline_thickness: int = 1,
        maxthickness_color: tuple[int, int, int] = (0, 255, 0),
        maxthickness_thickness: int = 1,
        uniformlayer_color: tuple[int, int, int] = (0, 0, 255),
        uniformlayer_thickness: int = 1,
        conformality_color: tuple[int, int, int] = (0, 0, 255),
        conformality_thickness: int = 1,
        conformality_step: int = 1,
        roughness_color: tuple[int, int, int] = (0, 0, 255),
        roughness_thickness: int = 1,
        roughness_step: int = 1,
    ) -> npt.NDArray[np.uint8]:
        """Visualize the analysis result.

        #. Draw the substrate with by :class:`PaintMode`.
        #. Display the template matching result with :class:`SubtractionMode`.
        #. Draw coating layer and contact line.
        #. If capillary bridge is broken, draw regional maximum thicknesses,
           uniform layer, conformality pairs and roughness pairs.

        Arguments:
            background_mode ({`'image', 'empty'`}): Determine how background is drawn.
                `'image'` draws original background image while `'empty'` draws on
                empty frame.
            subtraction_mode ({`'none', 'template', 'substrate', 'full'`}): Subtraction
                mode. `'template'` and `'substrate'` removes overlapping template region
                and substrate region, respectively. `'full'` removes both.
            layer_color: Layer contour's color for :func:`cv2.drawContours`.
            layer_thickness: Layer contour's thickness for :func:`cv2.drawContours`.
            contactline_color: Contact line's color for :func:`cv2.line`.
            contactline_thickness: Contact line's thickness for :func:`cv2.line`.
            maxthickness_color: Regional maximum thickness line's color for
                :func:`cv2.polylines`.
            maxthickness_thickness: Regional maximum thickness line's thickness for
                :func:`cv2.polylines`.
            uniformlayer_color: Imaginary uniform layer's color for
                :func:`cv2.polylines`.
            uniformlayer_thickness: Imaginary uniform layer's thickness for
                :func:`cv2.polylines`.
            conformality_color: Conformality pairs' color for :func:`cv2.polylines`.
            conformality_thickness: Conformality pairs' thickness for
                :func:`cv2.polylines`.
            conformality_step: Step size to skip conformality pairs.
            roughness_color: Roughness pairs' color for :func:`cv2.polylines`.
            roughness_thickness: Roughness pairs' thickness for :func:`cv2.polylines`.
            roughness_step: Step size to skip roughness pairs.
        """
        if background_mode == "image":
            image = self.image
        elif background_mode == "empty":
            image = np.full(self.image.shape, 255, dtype=np.uint8)
        else:
            raise TypeError("Unrecognized paint mode: %s" % background_mode)
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)  # type: ignore[assignment]

        if subtraction_mode not in ["none", "template", "substrate", "full"]:
            raise TypeError("Unrecognized subtraction mode: %s" % subtraction_mode)
        if subtraction_mode in ["template", "full"]:
            x0, y0, x1, y1 = self.substrate.reference.templateROI
            tempImg = self.substrate.reference.image[y0:y1, x0:x1]
            h, w = tempImg.shape[:2]
            (X0, Y0), _ = self.tempmatch
            binImg = self.image[Y0 : Y0 + h, X0 : X0 + w]
            mask = images_XOR(~binImg.astype(bool), ~tempImg.astype(bool))
            image[Y0 : Y0 + h, X0 : X0 + w][~mask] = 255
        if subtraction_mode in ["substrate", "full"]:
            x0, y0, x1, y1 = self.substrate.reference.substrateROI
            substImg = self.substrate.reference.image[y0:y1, x0:x1]
            h, w = substImg.shape[:2]
            X0, Y0 = self.substrate_point()
            binImg = self.image[Y0 : Y0 + h, X0 : X0 + w]
            mask = images_XOR(~binImg.astype(bool), ~substImg.astype(bool))
            image[Y0 : Y0 + h, X0 : X0 + w][~mask] = 255

        cv2.drawContours(image, self.layer_contours(), -1, layer_color, layer_thickness)

        if len(self.interfaces()) > 0:
            (i0, i1) = np.sort(np.concatenate(self.interfaces()).flatten())[[0, -1]]
            subst_cnt = self.substrate.contour() + self.substrate_point()
            (p0,), (p1,) = subst_cnt[[i0, i1]].astype(np.int32)
            cv2.line(image, p0, p1, contactline_color, contactline_thickness)

        if not self.capbridge_broken():
            return image

        lines = []
        for dist, pts in zip(*self.max_thickness()):
            if dist == 0:
                continue
            lines.append(pts.astype(np.int32))
        cv2.polylines(
            image,
            lines,
            isClosed=False,
            color=maxthickness_color,
            thickness=maxthickness_thickness,
        )

        _, points = self.uniform_layer()
        cv2.polylines(
            image,
            [points.astype(np.int32)],
            isClosed=False,
            color=uniformlayer_color,
            thickness=uniformlayer_thickness,
        )

        if len(self.interfaces()) > 0:
            I0, I1 = self.surface()
            surf = self.contour()[I0:I1]
            (i0, i1) = np.sort(np.concatenate(self.interfaces()).flatten())[[0, -1]]
            intf = (self.substrate.contour() + self.substrate_point())[i0:i1]
            _, path = self.conformality()
            path = path[::conformality_step]
            lines = np.concatenate([surf[path[..., 0]], intf[path[..., 1]]], axis=1)
            cv2.polylines(
                image,
                lines,
                isClosed=False,
                color=conformality_color,
                thickness=conformality_thickness,
            )

        if len(self.interfaces()) > 0:
            I0, I1 = self.surface()
            surf = self.contour()[I0:I1]
            _, ul = self.uniform_layer()
            ul_len = np.ceil(cv2.arcLength(ul.astype(np.float32), closed=False))
            ul = equidistant_interpolate(ul, int(ul_len))
            _, path = self.roughness()
            path = path[::roughness_step]
            lines = np.concatenate(
                [surf[path[..., 0]], ul[path[..., 1]]], axis=1
            ).astype(np.int32)
            cv2.polylines(
                image,
                lines,
                isClosed=False,
                color=roughness_color,
                thickness=roughness_thickness,
            )

        return image


def images_XOR(
    img1: npt.NDArray[np.bool_],
    img2: npt.NDArray[np.bool_],
    point: tuple[int, int] = (0, 0),
) -> npt.NDArray[np.bool_]:
    """Perform subtraction between two images by XOR operation.

    Arguments:
        img1: Image from which *img2* is subtracted.
        img2: Image patch which is subtracted from *img1*.
        point: Location in *img1* where *img2* is subtracted.
    """
    H, W = img1.shape
    h, w = img2.shape
    x0, y0 = point
    x1, y1 = x0 + w, y0 + h

    img1 = img1.copy()
    img1_crop = img1[max(y0, 0) : min(y1, H), max(x0, 0) : min(x1, W)]
    img2_crop = img2[max(-y0, 0) : min(H - y0, h), max(-x0, 0) : min(W - x0, w)]
    img1_crop ^= img2_crop
    return img1


def images_ANDXOR(
    img1: npt.NDArray[np.bool_],
    img2: npt.NDArray[np.bool_],
    point: tuple[int, int] = (0, 0),
) -> npt.NDArray[np.bool_]:
    """Perform subtraction between two images by AND and XOR operations.

    Arguments:
        img1: Image from which *img2* is subtracted.
        img2: Image patch which is subtracted from *img1*.
        point: Location in *img1* where *img2* is subtracted.
    """
    H, W = img1.shape
    h, w = img2.shape
    x0, y0 = point
    x1, y1 = x0 + w, y0 + h

    img1 = img1.copy()
    img1_crop = img1[max(y0, 0) : min(y1, H), max(x0, 0) : min(x1, W)]
    img2_crop = img2[max(-y0, 0) : min(H - y0, h), max(-x0, 0) : min(W - x0, w)]
    common = img1_crop & img2_crop
    img1_crop ^= common
    return img1


def equidistant_interpolate(points, n) -> npt.NDArray[np.float64]:
    """Interpolate points with equidistant new points.

    Arguments:
        points: Points that are interpolated.
            The shape must be ``(N, 1, D)`` where ``N`` is the number of points
            and ``D`` is the dimension.
        n: Number of new points.

    Returns:
        Interpolated points. If ``N`` is positive number, the shape is ``(n, 1, D)``.
        If ``N`` is zero, the shape is ``(n, 0, D)``.
    """
    # https://stackoverflow.com/a/19122075
    if points.size == 0:
        return np.empty((n, 0, points.shape[-1]), dtype=np.float64)
    vec = np.diff(points, axis=0)
    dist = np.linalg.norm(vec, axis=-1)
    u = np.insert(np.cumsum(dist), 0, 0)
    t = np.linspace(0, u[-1], n)
    ret = np.column_stack([np.interp(t, u, a) for a in np.squeeze(points, axis=1).T])
    return ret.reshape((n,) + points.shape[1:])


def parallel_curve(curve: npt.NDArray, dist: float) -> npt.NDArray:
    """Return parallel curve of *curve* with offset distance *dist*.

    Arguments:
        curve: Vertices of a polyline.
            The shape is ``(V, 1, D)``, where ``V`` is the number of vertices and
            ``D`` is the dimension.
        dist: offset distance of the parallel curve.

    Returns:
        Round-joint parallel curve of shape ``(V, 1, D)``.
    """
    if dist == 0:
        return curve
    ret = offset_curve(LineString(np.squeeze(curve, axis=1)), dist, join_style="round")
    return np.array(ret.coords)[:, np.newaxis]


@njit(cache=True)
def acm(cm: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """Compute accumulated cost matrix from local cost matrix.

    Arguments:
        cm: Local cost matrix.

    Returns:
        Accumulated cost matrix. The element at `[-1, -1]` is the total sum along the
        optimal path. If *cm* is empty, return value is an empty array.
    """
    p, q = cm.shape
    ret = np.zeros((p, q), dtype=np.float64)
    if p == 0 or q == 0:
        return ret

    ret[0, 0] = cm[0, 0]

    for i in range(1, p):
        ret[i, 0] = ret[i - 1, 0] + cm[i, 0]

    for j in range(1, q):
        ret[0, j] = ret[0, j - 1] + cm[0, j]

    for i in range(1, p):
        for j in range(1, q):
            ret[i, j] = min(ret[i - 1, j], ret[i, j - 1], ret[i - 1, j - 1]) + cm[i, j]

    return ret


@njit(cache=True)
def owp(acm: npt.NDArray[np.float64]) -> npt.NDArray[np.int32]:
    """Compute optimal warping path from accumulated cost matrix.

    Arguments:
        acm: Accumulated cost matrix.

    Returns:
        Indices for the two series to get the optimal warping path.
    """
    p, q = acm.shape
    if p == 0 or q == 0:
        return np.empty((0, 2), dtype=np.int32)

    path = np.zeros((p + q - 1, 2), dtype=np.int32)
    path_len = np.int32(0)

    i, j = p - 1, q - 1
    path[path_len] = [i, j]
    path_len += np.int32(1)

    while i > 0 or j > 0:
        if i == 0:
            j -= 1
        elif j == 0:
            i -= 1
        else:
            d = min(acm[i - 1, j], acm[i, j - 1], acm[i - 1, j - 1])
            if acm[i - 1, j] == d:
                i -= 1
            elif acm[i, j - 1] == d:
                j -= 1
            else:
                i -= 1
                j -= 1

        path[path_len] = [i, j]
        path_len += np.int32(1)

    return path[-(len(path) - path_len + 1) :: -1, :]



"""Analyze substrate geometry."""

import abc
import dataclasses
from typing import TYPE_CHECKING, Generic, TypeVar

import cv2
import numpy as np
import numpy.typing as npt
from numba import njit  # type: ignore
from scipy.ndimage import gaussian_filter1d  # type: ignore
from scipy.signal import find_peaks, peak_prominences  # type: ignore

from .cache import attrcache
from .reference import ReferenceBase

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

__all__ = [
    "RefTypeVar",
    "DataTypeVar",
    "SubstrateBase",
    "SubstrateData",
    "Substrate",
    "PolySubstrateBase",
    "houghline_accum",
    "RectSubstData",
    "RectSubstrate",
]


RefTypeVar = TypeVar("RefTypeVar", bound=ReferenceBase)
"""Type variable for the reference type of :class:`SubstrateBase`."""
DataTypeVar = TypeVar("DataTypeVar", bound="DataclassInstance")
"""Type variable for :attr:`SubstrateBase.DataType`."""


class SubstrateBase(abc.ABC, Generic[RefTypeVar, DataTypeVar]):
    """Abstract base class for substrate object.

    Substrate object stores substrate image, which is a binary image of bare substrate
    acquired from :class:`ReferenceBase` object. The role of substrate object is to
    analyze the shape of the bare substrate.

    External API can use the following members to get analysis results of
    concrete subclasses.

    * :attr:`DataType`: Dataclass type for the analysis result.
    * :meth:`analyze`: :attr:`DataType` instance containing analysis result.
    * :meth:`draw`: Visualized result.

    Arguments:
        reference: Reference instance which contains the substrate image.
    """

    DataType: type[DataTypeVar]
    """Return type of :attr:`analyze`.

    Concrete subclass must assign this attribute with dataclass type.
    """

    def __init__(self, reference: RefTypeVar):
        """Initialize the instance."""
        self._reference = reference

    @property
    def reference(self) -> RefTypeVar:
        """Reference instance which contains the substrate image."""
        return self._reference

    def image(self) -> npt.NDArray[np.uint8]:
        """Substrate image from :meth:`reference`."""
        x0, y0, x1, y1 = self.reference.substrateROI
        return self.reference.image[y0:y1, x0:x1]

    @abc.abstractmethod
    def region_points(self) -> npt.NDArray[np.int32]:
        """Coordinates of points representing each substrate region.

        Substrate image can have multiple disconnected substrate regions.
        Concrete classes should implement this method to return coordinates of
        points representing each region.

        Returns:
            `(N, 2)`-shaped array, where `N` is the number of substrate regions.
            Column should be the coordinates of points in ``(x, y)``.

        Note:
            These points are used to distinguish substrate regions from other
            foreground pixels, and give indices to each region.

            As higher-level methods are expected to rely on this method,
            it is best to keep this method simple and independent.
        """

    @attrcache("_regions")
    def regions(self) -> npt.NDArray[np.int8]:
        """Labelled image of substrate regions.

        Substrate regions are determined as connected component including
        a point in :meth:`region_points`.

        Returns:
            Labelled image.
            Value of ``i`` represents ``i``-th region in :meth:`region_points`.
            ``-1`` represents background.

        Note:
            Maximum number of regions is 128.
        """
        ret = np.full(self.image().shape[:2], -1, dtype=np.int8)
        _, labels = cv2.connectedComponents(cv2.bitwise_not(self.image()))
        for i, pt in enumerate(self.region_points()):
            ret[labels == labels[pt[1], pt[0]]] = i
        return ret

    def contours(
        self, region: int
    ) -> tuple[tuple[npt.NDArray[np.int32], ...], npt.NDArray[np.int32]]:
        """Find contours of a substrate region.

        Arguments:
            region: Label of the region from :meth:`regions`.

        Returns:
            Tuple of the result of :func:`cv2.findContours`.

        Note:
            Contours are dense, i.e., no approximation is made.
        """
        reg = (self.regions() == region) * np.uint8(255)
        return cv2.findContours(reg, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)

    @abc.abstractmethod
    def analyze(self) -> DataTypeVar:
        """Return analysis result as dataclass.

        Return type must be :attr:`DataType`.
        """

    @abc.abstractmethod
    def draw(self, *args, **kwargs) -> npt.NDArray[np.uint8]:
        """Return visualization result."""


@dataclasses.dataclass
class SubstrateData:
    """Analysis data for :class:`Substrate`."""

    pass


class Substrate(SubstrateBase[ReferenceBase, SubstrateData]):
    """Basic implementation of substrate without any geometric specification.

    Arguments:
        reference: Reference instance which contains the substrate image.

    Examples:
        .. plot::
            :include-source:
            :context: reset

            >>> import cv2
            >>> from finitedepth import get_sample_path, Reference, Substrate
            >>> img = cv2.imread(get_sample_path("ref.png"), cv2.IMREAD_GRAYSCALE)
            >>> _, bin = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
            >>> ref = Reference(bin, (10, 10, 1250, 200), (100, 100, 1200, 500))
            >>> subst = Substrate(ref)
            >>> import matplotlib.pyplot as plt #doctest: +SKIP
            >>> plt.imshow(subst.draw()) #doctest: +SKIP
    """

    DataType = SubstrateData
    """Return :obj:`SubstrateData`."""

    def region_points(self) -> npt.NDArray[np.int32]:
        """Return an upper center point of the substrate image.

        Substrate ROI in reference image must be selected so that this point falls into
        substrate region.
        """
        return np.array([[self.image().shape[1] / 2, 0]], dtype=np.int32)

    def analyze(self):
        """Return empty :class:`SubstrateData`."""
        return self.DataType()

    def draw(self) -> npt.NDArray[np.uint8]:
        """Return :meth:`image` in RGB."""
        ret = cv2.cvtColor(self.image(), cv2.COLOR_GRAY2RGB)
        return ret  # type: ignore[return-value]


class PolySubstrateBase(SubstrateBase[RefTypeVar, DataTypeVar]):
    """Abstract base class for substrate whose cross section is a simple polygon.

    A simple polygon does not have intersection nor hole [#poly]_. Smooth corners are
    allowed.

    Arguments:
        reference: Reference instance which contains the substrate image.

    Note:
        Substrate image should *not* have:
            * Multiple substrates in one image
            * Multiple contours (e.g. substrate with holes)

    .. [#poly] https://en.wikipedia.org/wiki/Simple_polygon
    """

    @property
    @abc.abstractmethod
    def sigma(self) -> float:
        """Standard deviation of gaussian filter to smooth the noise in contour."""

    @property
    @abc.abstractmethod
    def hough_parameters(self) -> tuple[float, float, int]:
        r"""Parameters for Hough line transformation.

        Returns:
            Tuple of numbers.

            * rho: Resolution of :math:`\rho`.
            * theta: Resolution of :math:`\theta`.
            * step: Step size to jump the points for better performance.
        """

    @abc.abstractmethod
    def n(self) -> int:
        """Number of polygon vertices."""

    def region_points(self) -> npt.NDArray[np.int32]:
        """Return an upper center point of the substrate image.

        Substrate ROI in reference image must be selected so that this point falls into
        substrate region.
        """
        return np.array([[self.image().shape[1] / 2, 0]], dtype=np.int32)

    def contour(self) -> npt.NDArray[np.int32]:
        """Return the polygon contour."""
        (cnt,), _ = self.contours(0)
        return cnt

    def vertices(self) -> npt.NDArray[np.int32]:
        """Find :meth:`n` vertices of the polygon.

        A vertex is a point where two sides of a polygon meet [#vertex-geom]_.
        When the sides are curves, the vertices are defined as local extrema of
        curvature [#vertex-curve]_.

        The vertices are found by smoothing :meth:`contour` with :attr:`sigma` and
        finding the local extrema of curvature [#contour-curvature]_.

        Returns:
            Indices of the vertex points in :meth:`contour`.

        .. [#vertex-geom] https://en.wikipedia.org/wiki/Vertex_(geometry)
        .. [#vertex-curve] https://en.wikipedia.org/wiki/Vertex_(curve)
        .. [#contour-curvature] https://stackoverflow.com/q/32629806
        """
        cnt = self.contour().astype(np.float64)

        # 1. Calculate curvatures
        f = gaussian_filter1d(cnt, self.sigma, axis=0, order=0, mode="wrap")
        f_dt = np.gradient(f, axis=0)
        f_dt2 = np.gradient(f_dt, axis=0)
        K = np.abs(np.cross(f_dt, f_dt2)) / np.linalg.norm(f_dt, axis=-1) ** 3

        # 2. Repeat the array (periodic)
        L = len(K)
        (K_rpt,) = np.concatenate([K[-L // 2 :], K, K[: L // 2]], axis=0).T

        # 3. Find peak
        peaks = find_peaks(K_rpt)[0].astype(np.int32)
        (idxs,) = np.where((L // 2 <= peaks) & (peaks < 3 * L // 2))
        peaks = peaks[idxs]
        prom, _, _ = peak_prominences(K_rpt, peaks)
        prom_peaks = peaks[np.argsort(prom)[-self.n() :]]
        return np.sort((prom_peaks - (L // 2)) % L)

    def sides(self) -> tuple[npt.NDArray[np.int32], ...]:
        """Find :meth:`n` sides of the polygon.

        The sides are found by slicing :meth:`contour` by :meth:`vertices`.

        Returns:
            Tuple of array containing points on each side of the polygon contour.
            The arrays are sorted so that the side containing the first point of the
            contour comes first.

        Note:
            Sides can be noisy and curved. Use :meth:`sidelines` to get linear models.

            The term "side" is used instead of "edge" to avoid confusion from
            other image processing methods (e.g. Canny edge detection).
        """
        cnt = self.contour()
        vert = self.vertices()
        sides = np.split(cnt, vert)
        dists = np.linalg.norm(cnt[vert] - cnt[0], axis=-1)
        shift = vert[np.argmin(dists)]
        L = len(cnt)
        sides = np.split(np.roll(cnt, shift, axis=0), np.sort((vert - shift) % L))[1:]
        return tuple(sides)

    @attrcache("_sidelines")
    def sidelines(self) -> npt.NDArray[np.float32]:
        r"""Find :meth:`n` sidelines of the polygon.

        Sideline is the line that contains one side of the polygon [#sideline]_.
        The sidelines are found by performing Hough line transformation on :meth:`sides`
        with :attr:`hough_parameters`.

        Returns:
            Vector of line parameters in :math:`(\rho, \theta)`.
            :math:`\rho` is the distance from the coordinate origin.
            :math:`\theta` is the angle of normal vector from the origin
            to the line.

        Note:
            Range of angle is
            :math:`\theta \in (-\frac{3 \pi}{2}, \frac{\pi}{2}]`.
            Arctangent direction can be acquired by
            :math:`\theta + \frac{\pi}{2}`.

        .. [#sideline] https://en.wikipedia.org/wiki/Extended_side
        """
        # Do not find the line from smoothed contour. Noise is removed anyway
        # without smoothing by Hough transformation. In fact, smoothing
        # propagates the outlier error to nearby data.
        RHO_RES, THETA_RES, STEP = self.hough_parameters
        lines = []
        # Directly use Hough transformation to find lines
        for side in self.sides():
            tan = np.diff(side, axis=0)
            atan = np.arctan2(tan[..., 1], tan[..., 0])  # -pi < atan <= pi
            theta = atan - np.pi / 2
            tmin, tmax = theta.min(), theta.max()
            if tmin < tmax:
                theta_rng = np.arange(tmin, tmax, THETA_RES, dtype=np.float32)
            else:
                theta_rng = np.array([tmin], dtype=np.float32)

            # Interpolate & perform hough transformation.
            c = side[::STEP]
            rho = c[..., 0] * np.cos(theta_rng) + c[..., 1] * np.sin(theta_rng)
            rho_digit = (rho / RHO_RES).astype(np.int32)

            _, (rho, theta_idx) = houghline_accum(rho_digit)
            lines.append([[rho * RHO_RES, theta_rng[theta_idx]]])

        return np.array(lines, dtype=np.float32)

    def sideline_intersections(self) -> npt.NDArray[np.float32]:
        """Find intersections of :meth:`sidelines`."""
        ((r1, t1),) = self.sidelines().transpose(1, 2, 0)
        ((r2, t2),) = np.roll(self.sidelines(), 1, axis=0).transpose(1, 2, 0)
        mat = np.array([[np.cos(t1), np.sin(t1)], [np.cos(t2), np.sin(t2)]]).transpose(
            2, 0, 1
        )
        vec = np.array([[r1], [r2]]).transpose(2, 0, 1)
        (ret,) = (np.linalg.inv(mat) @ vec).transpose(2, 0, 1)
        return ret


@njit(cache=True)
def houghline_accum(
    rho_array: npt.NDArray[np.int32],
) -> tuple[npt.NDArray[np.int32], tuple[float, int]]:
    """Perform hough line accumulation.

    Arguments:
        rho_array: Array containing rho and theta values for every points.
            The shape must be ``(P, T)``, where ``P`` is the number of points
            and ``T`` is the numbers of digitized theta intervals.

    Returns:
        Tuple of accumulation matrix and detected ``(rho, theta_idx)`` value.
    """
    rho_min = np.min(rho_array)
    n_rho = np.max(rho_array) - rho_min + 1
    n_pts, n_theta = rho_array.shape
    accum = np.zeros((n_rho, n_theta), dtype=np.int32)

    maxloc = (0, 0)
    for i in range(n_pts):
        for j in range(n_theta):
            r = rho_array[i, j] - rho_min
            accum[r, j] += 1
            if accum[r, j] > accum[maxloc]:
                maxloc = (r, j)

    rho_theta = (float(maxloc[0] + rho_min), int(maxloc[1]))
    return accum, rho_theta


@dataclasses.dataclass
class RectSubstData:
    """Analysis data for :class:`RectSubstrate`.

    Arguments:
        Width: Width of the rectangular cross section in pixels.
    """

    Width: np.float32


class RectSubstrate(PolySubstrateBase[ReferenceBase, RectSubstData]):
    """Substrate having rectangular cross section.

    Arguments:
        reference: Reference instance which contains the substrate image.
        sigma: Standard deviation of gaussian filter to smooth the noise in contour.
        rho_thres, theta_thres, hough_step: Hough line transformation parameters.

    Examples:
        .. plot::
            :include-source:
            :context: reset

            >>> import cv2
            >>> from finitedepth import get_sample_path, Reference, RectSubstrate
            >>> img = cv2.imread(get_sample_path("ref.png"), cv2.IMREAD_GRAYSCALE)
            >>> _, bin = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
            >>> ref = Reference(bin, (10, 10, 1250, 200), (100, 100, 1200, 500))
            >>> subst = RectSubstrate(ref, 3.0, 1.0, 0.01)
            >>> import matplotlib.pyplot as plt #doctest: +SKIP
            >>> plt.imshow(subst.draw()) #doctest: +SKIP
    """

    DataType = RectSubstData
    """Return :obj:`RectSubstData`."""

    def __init__(
        self,
        reference: ReferenceBase,
        sigma: float,
        rho_thres: float,
        theta_thres: float,
        hough_step: int = 1,
    ):
        """Initialize the instance by passed arguments."""
        super().__init__(reference)
        self._sigma = sigma
        self._hough_param = (rho_thres, theta_thres, hough_step)

    @property
    def sigma(self) -> float:
        """Sigma value passed to the constructor."""
        return self._sigma

    @property
    def hough_parameters(self) -> tuple[float, float, int]:
        """Hough line transformation parameters passed to the constructor."""
        return self._hough_param

    def n(self) -> int:
        """Number of vertices of rectangle, which is 4."""
        return 4

    def analyze(self):
        """Return :class:`RectSubstData`."""
        _, B, C, _ = self.sideline_intersections()
        return self.DataType(np.linalg.norm(B - C))

    def draw(
        self,
        mode: str = "image",
        vertice_color: tuple[int, int, int] = (0, 255, 0),
        vertice_thickness: int = 1,
        vertice_markerSize: int = 20,
        sideline_color: tuple[int, int, int] = (0, 0, 255),
        sideline_thickness: int = 1,
    ) -> npt.NDArray[np.uint8]:
        """Draw substrate image and show vertices and sidelines.

        Arguments:
            mode ({`'image', 'contour'`}): Draw mode.
                `'image'` draws :meth:`image`, while `'contour'` draws :meth:`contour`.
            vertice_color: Vertice marker color for :func:`cv2.drawMarker`.
            vertice_thickness: Vertice marker thickness for :func:`cv2.drawMarker`.
            vertice_markerSize: Vertice marker size for :func:`cv2.drawMarker`.
            sideline_color: Sideline color for :func:`cv2.line`.
            sideline_thickness: Sideline thickness for :func:`cv2.line`.
        """
        if mode == "image":
            image = self.image()
        elif mode == "contour":
            h, w = self.image().shape[:2]
            image = np.full((h, w), 255, np.uint8)
            cv2.drawContours(image, self.contour(), -1, 0, 1)  # type: ignore
        else:
            raise TypeError("Unrecognized draw mode: %s" % mode)
        ret = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)

        if vertice_thickness > 0:
            for (pt,) in self.contour()[self.vertices()]:
                cv2.drawMarker(
                    ret,
                    pt,
                    color=vertice_color,
                    markerType=cv2.MARKER_CROSS,
                    markerSize=vertice_markerSize,
                    thickness=vertice_thickness,
                )

        if sideline_thickness > 0:
            try:
                tl, bl, br, tr = self.sideline_intersections().astype(np.int32)

                cv2.line(ret, tl, tr, sideline_color, sideline_thickness)
                cv2.line(ret, tr, br, sideline_color, sideline_thickness)
                cv2.line(ret, br, bl, sideline_color, sideline_thickness)
                cv2.line(ret, bl, tl, sideline_color, sideline_thickness)
            except ValueError:
                pass

        return ret  # type: ignore[return-value]



"""Package for image analysis of finite depth dip coating.

To analyze with command line, specify the parameters in configuration file(s) and run::

    finitedepth analyze <file1> [<file2> ...]
"""

import argparse
import csv
import dataclasses
import glob
import json
import logging
import os
import re
import sys
from importlib.metadata import entry_points
from importlib.resources import files

import cv2
import tqdm  # type: ignore
import yaml

from .coatinglayer import CoatingLayer, CoatingLayerBase, RectLayerShape
from .reference import Reference, ReferenceBase
from .substrate import PolySubstrateBase, RectSubstrate, Substrate, SubstrateBase

__all__ = [
    "get_sample_path",
    "analyze_files",
    "ReferenceBase",
    "Reference",
    "SubstrateBase",
    "Substrate",
    "PolySubstrateBase",
    "RectSubstrate",
    "CoatingLayerBase",
    "CoatingLayer",
    "RectLayerShape",
]


def get_sample_path(*paths: str) -> str:
    """Get path to sample file.

    Arguments:
        paths: Subpaths under ``finitedepth/samples/`` directory.

    Returns:
        Absolute path to the sample file.

    Examples:
        >>> from finitedepth import get_sample_path
        >>> get_sample_path() # doctest: +SKIP
        'path/finitedepth/samples'
        >>> get_sample_path("myfile") # doctest: +SKIP
        'path/finitedepth/samples/myfile'
    """
    return str(files("finitedepth").joinpath("samples", *paths))


def analyze_files(
    *paths: str, recursive: bool = False, entries: list[str] | None = None
) -> bool:
    """Perform analysis from configuration files.

    Supported formats:
        * YAML
        * JSON

    Each file can have multiple entries. Each entry must have ``type`` field which
    specifies the analyzer. For example, the following YAML file contains ``foo``
    entry which is analyzed by ``Foo`` analyzer.

    .. code-block:: yaml

        foo:
            type: Foo
            ...

    Analyzers are searched and loaded from entry point group
    ``"finitedepth.analyzers"``, and must have the following signature:

    * entry name (:obj:`str`)
    * entry fields (:obj:`dict`)

    Arguments:
        paths: Glob pattern for configuration file paths.
        recursive: If True, search *paths* recursively.
        entries: Regular expression for entries.
            If passed, only the matching entries are analyzed.

    Returns:
        Whether the analysis is finished without error.
    """
    # load analyzers
    ANALYZERS = {}
    for ep in entry_points(group="finitedepth.analyzers"):
        ANALYZERS[ep.name] = ep

    glob_paths = []
    for path in paths:
        glob_paths.extend(glob.glob(os.path.expandvars(path), recursive=recursive))

    if entries is not None:
        entry_patterns = [re.compile(e) for e in entries]
    else:
        entry_patterns = []

    ok = True
    for path in glob_paths:
        _, ext = os.path.splitext(path)
        ext = ext.lstrip(os.path.extsep).lower()
        try:
            with open(path, "r") as f:
                if ext == "yaml" or ext == "yml":
                    data = yaml.load(f, Loader=yaml.FullLoader)
                elif ext == "json":
                    data = json.load(f)
                else:
                    logging.error(f"Skipping file: '{path}' (format not supported)")
                    ok = False
                    continue
        except FileNotFoundError:
            logging.error(f"Skipping file: '{path}' (does not exist)")
            ok = False
            continue
        for k, v in data.items():
            if entry_patterns and all([p.fullmatch(k) is None for p in entry_patterns]):
                continue
            try:
                typename = v["type"]
                analyzer = ANALYZERS.get(typename, None)
                if analyzer is not None:
                    analyzer.load()(k, v)
                else:
                    logging.error(
                        f"Skipping entry: '{path}::{k}' (unknown type: '{typename}')"
                    )
            except Exception:
                logging.exception(f"Skipping entry: '{path}::{k}' (exception raised)")
                ok = False
                continue
    return ok


def coatingimage_analyzer(name, data):
    """Image analysis for a single image of coated substrate.

    Coating layer is analyzed by constructing :class:`CoatingLayerBase` implementation.
    The analyzer defines the following fields in configuration entry:

    - **referencePath** (`str`): Path to the reference image file.
        Image is converted to binary by Otsu's thresholding.
    - **targetPath** (`str`): Path to the target image file.
        Image is converted to binary by Otsu's thresholding.
    - **reference**, **substrate**, **layer** (`mapping`, optional):
        - **type** (`str`, optional): Registered class constructor.
          The constructors are found from entry point groups
          `'finitedepth.(references|substrates|coatinglayers)'`.
        - **parameters** (`mapping`, optional): Any additional parameter
          for class constructor.
        - **draw** (`mapping`, optional): Parameters for :meth:`draw`.
    - **output** (`mapping`, optional):
        - **(reference|substrate|layer)Data** (`str`, optional):
          Path to the output CSV file containing analysis result.
          The results are acquired from :meth:`analyze`.
        - **(reference|substrate|layer)Image** (`str`, optional):
          Path to the output image file containing visualization result.
          The results are acquired from :meth:`draw`.

    The default constructors used when **type** is not specified are :class:`Reference`,
    :class:`Substrate`, and :class:`CoatingLayer`.

    The following is the example for an entry in YAML configuration file:

    .. code-block:: yaml

        foo:
            type: CoatingImage
            referencePath: foo-ref.png
            targetPath: foo-target.png
            reference:
                parameters:
                    templateROI: [10, 10, 1250, 200]
                    substrateROI: [100, 100, 1200, 500]
            layer:
                draw:
                    layer_color: [0, 255, 0]
            output:
                layerImage: output/foo.jpg
                layerData: output/foo.csv
    """
    output = data.get("output", {})
    output_refdata = _makedir(output.get("referenceData", ""))
    output_refimg = _makedir(output.get("referenceImage", ""))
    output_substdata = _makedir(output.get("substrateData", ""))
    output_substimg = _makedir(output.get("substrateImage", ""))
    output_layerdata = _makedir(output.get("layerData", ""))
    output_layerimg = _makedir(output.get("layerImage", ""))

    RefType, SubstType, LayerType = _load_types(data)

    if output_refdata:
        csvwriter = _CsvWriter(output_refdata)
        next(csvwriter)

    try:
        refdata = data.get("reference", {})
        for path in tqdm.tqdm([data["referencePath"]], desc=f"{name} (ref)"):
            _, refimg = cv2.threshold(
                cv2.imread(os.path.expandvars(path), cv2.IMREAD_GRAYSCALE),
                0,
                255,
                cv2.THRESH_BINARY | cv2.THRESH_OTSU,
            )
            ref = RefType(refimg, **refdata.get("parameters", {}))
            if output_refdata:
                csvwriter.send([d.name for d in dataclasses.fields(ref.DataType)])
                csvwriter.send(dataclasses.astuple(ref.analyze()))
            if output_refimg:
                cv2.imwrite(
                    output_refimg,
                    cv2.cvtColor(
                        ref.draw(**refdata.get("draw", {})), cv2.COLOR_BGR2RGB
                    ),
                )
    finally:
        if output_refdata:
            csvwriter.close()

    if output_substdata:
        csvwriter = _CsvWriter(output_substdata)
        next(csvwriter)

    try:
        substdata = data.get("substrate", {})
        for reference in tqdm.tqdm([ref], desc=f"{name} (subst)"):
            subst = SubstType(reference, **substdata.get("parameters", {}))
            if output_substdata:
                csvwriter.send([d.name for d in dataclasses.fields(subst.DataType)])
                csvwriter.send(dataclasses.astuple(subst.analyze()))
            if output_substimg:
                cv2.imwrite(
                    output_substimg,
                    cv2.cvtColor(
                        subst.draw(**substdata.get("draw", {})), cv2.COLOR_BGR2RGB
                    ),
                )
    finally:
        if output_substdata:
            csvwriter.close()

    if output_layerdata:
        csvwriter = _CsvWriter(output_layerdata)
        next(csvwriter)

    try:
        layerdata = data.get("layer", {})
        for path in tqdm.tqdm([data["targetPath"]], desc=f"{name} (layer)"):
            _, tgtimg = cv2.threshold(
                cv2.imread(os.path.expandvars(path), cv2.IMREAD_GRAYSCALE),
                0,
                255,
                cv2.THRESH_BINARY | cv2.THRESH_OTSU,
            )
            layer = LayerType(tgtimg, subst, **layerdata.get("parameters", {}))
            if output_layerdata:
                csvwriter.send([d.name for d in dataclasses.fields(layer.DataType)])
                csvwriter.send(dataclasses.astuple(layer.analyze()))
            if output_layerimg:
                cv2.imwrite(
                    output_layerimg,
                    cv2.cvtColor(
                        layer.draw(**layerdata.get("draw", {})), cv2.COLOR_BGR2RGB
                    ),
                )
    finally:
        if output_layerdata:
            csvwriter.close()


def coatingvideo_analyzer(name, data):
    """Image analysis for coated substrate video.

    Coating layer is analyzed by constructing :class:`CoatingLayerBase` implementation.
    The analyzer defines the following fields in configuration entry:

    - **referencePath** (`str`): Path to the reference image file.
        Image is converted to binary by Otsu's thresholding.
    - **targetPath** (`str`): Path to the target video file.
        Frames are converted to binary by Otsu's thresholding.
    - **reference**, **substrate**, **layer** (`mapping`, optional):
        - **type** (`str`, optional): Registered class constructor.
          The constructors are found from entry point groups
          `'finitedepth.(references|substrates|coatinglayers)'`.
        - **parameters** (`mapping`, optional): Any additional parameter
          for class constructor.
        - **draw** (`mapping`, optional): Parameters for :meth:`draw`.
    - **output** (`mapping`, optional):
        - **(reference|substrate|layer)Data** (`str`, optional):
          Path to the output CSV file containing analysis result.
          The results are acquired from :meth:`analyze`.
        - **(reference|substrate)Image** (`str`, optional):
          Path to the output image file containing visualization result.
          The results are acquired from :meth:`draw`.
        - **layerVideo** (`str`, optional):
          Path to the output video file containing visualization result.
          The frames are acquired from :meth:`draw`. The codec is set to be the same as
          the target video file.

    The default constructors used when **type** is not specified are :class:`Reference`,
    :class:`Substrate`, and :class:`CoatingLayer`.

    The following is the example for an entry in YAML configuration file:

    .. code-block:: yaml

        foo:
            type: CoatingImage
            referencePath: foo-ref.png
            targetPath: foo-target.mp4
            reference:
                parameters:
                    templateROI: [10, 10, 1250, 200]
                    substrateROI: [100, 100, 1200, 500]
            layer:
                draw:
                    layer_color: [0, 255, 0]
            output:
                layerVideo: output/foo.mp4
                layerData: output/foo.csv
    """
    output = data.get("output", {})
    output_refdata = _makedir(output.get("referenceData", ""))
    output_refimg = _makedir(output.get("referenceImage", ""))
    output_substdata = _makedir(output.get("substrateData", ""))
    output_substimg = _makedir(output.get("substrateImage", ""))
    output_layerdata = _makedir(output.get("layerData", ""))
    output_layervid = _makedir(output.get("layerVideo", ""))

    RefType, SubstType, LayerType = _load_types(data)

    if output_refdata:
        csvwriter = _CsvWriter(output_refdata)
        next(csvwriter)

    try:
        refdata = data.get("reference", {})
        for path in tqdm.tqdm([data["referencePath"]], desc=f"{name} (ref)"):
            _, refimg = cv2.threshold(
                cv2.imread(os.path.expandvars(path), cv2.IMREAD_GRAYSCALE),
                0,
                255,
                cv2.THRESH_BINARY | cv2.THRESH_OTSU,
            )
            ref = RefType(refimg, **refdata.get("parameters", {}))
            if output_refdata:
                csvwriter.send([d.name for d in dataclasses.fields(ref.DataType)])
                csvwriter.send(dataclasses.astuple(ref.analyze()))
            if output_refimg:
                cv2.imwrite(
                    output_refimg,
                    cv2.cvtColor(
                        ref.draw(**refdata.get("draw", {})), cv2.COLOR_BGR2RGB
                    ),
                )
    finally:
        if output_refdata:
            csvwriter.close()

    if output_substdata:
        csvwriter = _CsvWriter(output_substdata)
        next(csvwriter)

    try:
        substdata = data.get("substrate", {})
        for reference in tqdm.tqdm([ref], desc=f"{name} (subst)"):
            subst = SubstType(reference, **substdata.get("parameters", {}))
            if output_substdata:
                csvwriter.send([d.name for d in dataclasses.fields(subst.DataType)])
                csvwriter.send(dataclasses.astuple(subst.analyze()))
            if output_substimg:
                cv2.imwrite(
                    output_substimg,
                    cv2.cvtColor(
                        subst.draw(**substdata.get("draw", {})), cv2.COLOR_BGR2RGB
                    ),
                )
    finally:
        if output_substdata:
            csvwriter.close()

    if output_layerdata:
        csvwriter = _CsvWriter(output_layerdata)
        next(csvwriter)
    cap = cv2.VideoCapture(os.path.expandvars(data["targetPath"]))
    if output_layervid:
        vidwriter = cv2.VideoWriter(
            output_layervid,
            int(cap.get(cv2.CAP_PROP_FOURCC)),
            cap.get(cv2.CAP_PROP_FPS),
            (
                int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
                int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)),
            ),
        )

    try:
        layerdata = data.get("layer", {})
        for i in tqdm.tqdm(
            range(int(cap.get(cv2.CAP_PROP_FRAME_COUNT))), desc=f"{name} (layer)"
        ):
            ok, frame = cap.read()
            if not ok:
                break
            _, tgtimg = cv2.threshold(
                cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY),
                0,
                255,
                cv2.THRESH_BINARY | cv2.THRESH_OTSU,
            )
            layer = LayerType(tgtimg, subst, **layerdata.get("parameters", {}))
            if output_layerdata:
                if i == 0:
                    csvwriter.send([d.name for d in dataclasses.fields(layer.DataType)])
                csvwriter.send(dataclasses.astuple(layer.analyze()))
            if output_layervid:
                vidwriter.write(
                    cv2.cvtColor(
                        layer.draw(**layerdata.get("draw", {})), cv2.COLOR_BGR2RGB
                    )
                )
    finally:
        cap.release()
        if output_layerdata:
            csvwriter.close()
        if output_layervid:
            vidwriter.release()


def _makedir(path: str) -> str:
    path = os.path.expandvars(path)
    dirname, _ = os.path.split(path)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    return path


def _load_types(
    data: dict,
) -> tuple[type[ReferenceBase], type[SubstrateBase], type[CoatingLayerBase]]:
    REFERENCES = {}
    for ep in entry_points(group="finitedepth.references"):
        REFERENCES[ep.name] = ep
    SUBSTRATES = {}
    for ep in entry_points(group="finitedepth.substrates"):
        SUBSTRATES[ep.name] = ep
    LAYERS = {}
    for ep in entry_points(group="finitedepth.coatinglayers"):
        LAYERS[ep.name] = ep

    refdata = data.get("reference", {})
    reftype = refdata.get("type", "")
    if reftype:
        RefType = REFERENCES[reftype].load()
    else:
        RefType = Reference

    substdata = data.get("substrate", {})
    substtype = substdata.get("type", "")
    if substtype:
        SubstType = SUBSTRATES[substtype].load()
    else:
        SubstType = Substrate

    layerdata = data.get("layer", {})
    layertype = layerdata.get("type", "")
    if layertype:
        LayerType = LAYERS[layertype].load()
    else:
        LayerType = CoatingLayer

    return (RefType, SubstType, LayerType)


def _CsvWriter(path):
    with open(path, "w", newline="") as f:
        writer = csv.writer(f)
        while True:
            row = yield
            writer.writerow(row)


def main():
    """Entry point function."""
    parser = argparse.ArgumentParser(
        prog="finitedepth",
        description="Finite depth dip coating image analysis.",
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        help="set logging level",
    )
    subparsers = parser.add_subparsers(dest="command")

    samples = subparsers.add_parser(
        "samples",
        description="Show path to sample directory.",
        help="show path to sample directory",
    ).add_mutually_exclusive_group()
    samples.add_argument(
        "plugin",
        type=str,
        nargs="?",
        help="name of the plugin",
    )
    samples.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="list plugin names",
    )

    subparsers.add_parser(
        "references",
        description="List installed reference constructors.",
        help="list installed reference constructors",
    )
    subparsers.add_parser(
        "substrates",
        description="List installed substrate constructors.",
        help="list installed substrate constructors",
    )
    subparsers.add_parser(
        "coatinglayers",
        description="List installed coating layer constructors.",
        help="list installed coating layer constructors",
    )

    analyze = subparsers.add_parser(
        "analyze",
        description="Parse configuration files and analyze.",
        help="parse configuration files and analyze",
        epilog=(
            "Supported file formats: YAML, JSON.\n"
            "Refer to the package documentation for configuration file structure."
        ),
    )
    analyze.add_argument(
        "file", type=str, nargs="+", help="glob pattern for configuration files"
    )
    analyze.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="recursively find configuration files",
    )
    analyze.add_argument(
        "-e",
        "--entry",
        action="append",
        help="regex pattern for configuration file entries",
    )

    args = parser.parse_args()

    loglevel = args.log_level.upper()
    logging.basicConfig(
        format="[%(asctime)s] [%(levelname)8s] --- %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=loglevel,
    )

    logging.debug(f"Input command: {' '.join(sys.argv)}")

    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(1)
    elif args.command == "samples":
        if args.list:
            header = [("PLUGIN", "PATH")]
            paths = [
                (ep.name, ep.load()())
                for ep in entry_points(group="finitedepth.samples")
            ]
            col0_max = max(len(p[0]) for p in header + paths)
            space = 3
            for col0, col1 in header + paths:
                line = col0.ljust(col0_max) + " " * space + col1
                print(line)
        elif args.plugin is None:
            print(get_sample_path())
        else:
            for ep in entry_points(group="finitedepth.samples"):
                if ep.name == args.plugin:
                    getter = ep.load()
                    print(getter())
                    break
            else:
                logging.error(
                    f"Unknown plugin: '{args.plugin}' (use '-l' option to list plugins)"
                )
                sys.exit(1)
    elif args.command in ["references", "substrates", "coatinglayers"]:
        header = [("NAME", "PACKAGE")]
        models = [
            (ep.name, ep.value.split(".")[0])
            for ep in entry_points(group=f"finitedepth.{args.command}")
        ]
        col0_max = max(len(m[0]) for m in header + models)
        space = 3
        for col0, col1 in header + models:
            line = col0.ljust(col0_max) + " " * space + col1
            print(line)
    elif args.command == "analyze":
        ok = analyze_files(*args.file, recursive=args.recursive, entries=args.entry)
        if not ok:
            sys.exit(1)



"""Caching helper."""

import functools

__all__ = [
    "attrcache",
]


def attrcache(attrname: str):
    """Decorate a nullary method to cache the result with attribute.

    Method must not have any argument.

    Examples:
        >>> from finitedepth.cache import attrcache
        >>> class Foo:
        ...     @attrcache("_bar")
        ...     def bar(self):
        ...         print("spam")
        ...         return 1
        >>> f = Foo()
        >>> f.bar()
        spam
        1
        >>> f.bar()
        1
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(self):
            if not hasattr(self, attrname):
                setattr(self, attrname, func(self))
            return getattr(self, attrname)

        return wrapper

    return decorator



import os
import subprocess

import yaml

from finitedepth import get_sample_path


def test_coatingimage(tmp_path):
    imgpath = tmp_path / "layer.png"
    datapath = tmp_path / "layer.csv"
    config = dict(
        data1=dict(
            type="CoatingImage",
            referencePath=get_sample_path("ref.png"),
            targetPath=get_sample_path("coat.png"),
            output=dict(layerImage=str(imgpath), layerData=str(datapath)),
        )
    )
    path = tmp_path / "config.yml"
    with open(path, "w") as f:
        yaml.dump(config, f)
    code = subprocess.call(
        [
            "finitedepth",
            "analyze",
            path,
        ],
    )
    assert not code
    assert os.path.exists(imgpath)
    assert os.path.exists(datapath)


def test_coatingvideo(tmp_path):
    vidpath = tmp_path / "layer.mp4"
    datapath = tmp_path / "layer.csv"
    config = dict(
        data1=dict(
            type="CoatingVideo",
            referencePath=get_sample_path("ref.png"),
            targetPath=get_sample_path("coat.mp4"),
            output=dict(layerVideo=str(vidpath), layerData=str(datapath)),
        )
    )
    path = tmp_path / "config.yml"
    with open(path, "w") as f:
        yaml.dump(config, f)
    code = subprocess.call(
        [
            "finitedepth",
            "analyze",
            path,
        ],
    )
    assert not code
    assert os.path.exists(vidpath)
    assert os.path.exists(datapath)
